@@ -16,18 +16,23 @@
 #include "term-query.h"
 
 #include <fcntl.h>
-#include <string.h>  // for memmem()
+#include <string.h>  // NOLINT for memmem()
 #include <sys/ioctl.h>
 #include <sys/select.h>
 #include <termios.h>
 #include <unistd.h>
 
+#include <cctype>
+#include <cstdint>
 #include <cstdio>
 #include <cstdlib>
 #include <ctime>
 #include <functional>
 
 #include "timg-time.h"
+#include "utils.h"
+
+#define TERM_CSI "\033["
 
 namespace timg {
 static struct termios s_orig_terminal_setting;
@@ -39,6 +44,27 @@ static void clean_up_terminal() {
     s_tty_fd = -1;
 }
 
+// Global variable; ok for debug logging cause.
+static bool s_log_terminal_queries = false;
+void EnableTerminalQueryLogging(bool on) { s_log_terminal_queries = on; }
+
+// Debug print a message and c-escaped data.
+static void debug_data(FILE *f, const char *msg, const char *data, size_t len) {
+    fprintf(f, "\033[1m%s\033[0m'", msg);
+    for (const char *const end = data + len; data < end; ++data) {
+        if (*data == '\\') {
+            fprintf(f, "\\\\");
+        }
+        else if (*data < 0x20) {
+            fprintf(f, "\\%03o", *data);
+        }
+        else {
+            fprintf(f, "%c", *data);
+        }
+    }
+    fprintf(f, "'");
+}
+
 // Send "query" to terminal and wait for response to arrive within
 // "time_budget". Use "buffer" with "len" to store results.
 // Whenever new data arrives, the caller's "response_found_p" response finder
@@ -56,8 +82,10 @@ static const char *QueryTerminal(const char *query, char *const buffer,
     // Let's see if we have at least one file descriptor that is connected
     // to our terminal. We can then open that terminal directly RD/WR.
     const char *ttypath = nullptr;
-    for (int fd : {STDOUT_FILENO, STDERR_FILENO, STDIN_FILENO}) {
-        if (isatty(fd) && (ttypath = ttyname(fd)) != nullptr) break;
+    for (const int fd : {STDOUT_FILENO, STDERR_FILENO, STDIN_FILENO}) {
+        if (!isatty(fd)) continue;
+        ttypath = ttyname(fd);
+        if (ttypath != nullptr) break;
     }
     if (!ttypath) return nullptr;
     s_tty_fd = open(ttypath, O_RDWR);
@@ -77,9 +105,9 @@ static const char *QueryTerminal(const char *query, char *const buffer,
     raw_terminal_setting.c_iflag     = 0;
     raw_terminal_setting.c_lflag &= ~(ICANON | ECHO);
 
-    if (tcsetattr(s_tty_fd, TCSANOW, &raw_terminal_setting) != 0)
+    if (tcsetattr(s_tty_fd, TCSANOW, &raw_terminal_setting) != 0) {
         return nullptr;
-
+    }
     // No matter what happens exiting early for some reason, make sure we
     // leave the terminal in a good state.
     atexit(clean_up_terminal);
@@ -92,15 +120,17 @@ static const char *QueryTerminal(const char *query, char *const buffer,
     const char *found_pos     = nullptr;
     size_t available          = buflen - 1;  // Allow for nul termination.
     char *pos                 = buffer;
-    timg::Time now            = Time::Now();
-    const timg::Time deadline = now + time_budget;
+    const timg::Time start    = Time::Now();
+    const timg::Time deadline = start + time_budget;
+    timg::Time now            = start;
     do {
         struct timeval timeout = (deadline - now).AsTimeval();
         fd_set read_fds;
         FD_ZERO(&read_fds);
         FD_SET(s_tty_fd, &read_fds);
-        if (select(s_tty_fd + 1, &read_fds, nullptr, nullptr, &timeout) <= 0)
+        if (select(s_tty_fd + 1, &read_fds, nullptr, nullptr, &timeout) <= 0) {
             break;
+        }
         const int r = read(s_tty_fd, pos, available);
         if (r < 0) break;
         pos += r;
@@ -113,10 +143,23 @@ static const char *QueryTerminal(const char *query, char *const buffer,
     } while (available && now < deadline);
 
     clean_up_terminal();
-
+    if (s_log_terminal_queries) {
+        debug_data(stderr, "Query: ", query, query_len);
+        debug_data(stderr, " Response: ", buffer, pos - buffer);
+        fprintf(stderr, " (%dms)\n",
+                (int)((timg::Time() - start).nanoseconds() / 1'000'000));
+    }
     return found_pos;
 }
 
+// Return pointer to substring in "haystack" of known length. The substring
+// "s" we look for is nul-terminated.
+// Not found is nullptr/false; if found, returns the start of the match.
+static const char *find_str(const char *haystack, int64_t len, const char *s) {
+    if (len < 0) return nullptr;
+    return (const char *)memmem(haystack, len, s, strlen(s));
+}
+
 // Read background color queried from terminal emulator.
 // Might leak a file-descriptor when bailing out early. Accepted for brevity.
 const char *QueryBackgroundColor() {
@@ -135,7 +178,6 @@ const char *QueryBackgroundColor() {
     const Duration kTimeBudget = Duration::Millis(1500);
 
     constexpr char query_background_color[] = "\033]11;?\033\\";
-    constexpr size_t kColorLen = 18;  // strlen("rgb:1234/1234/1234")
 
     // Query and testing the response. It is the query-string with the
     // question mark replaced with the requested information.
@@ -154,39 +196,39 @@ const char *QueryBackgroundColor() {
     //    Make sure to accumulate reads in a more spacious buffer than the
     //    expected response and finish once we found what we're looking for.
     char buffer[512];
-    const char *const start_color = QueryTerminal(
+    const char *start_color = QueryTerminal(
         query_background_color, buffer, sizeof(buffer), kTimeBudget,
         [](const char *data, size_t len) -> const char * {
             // We might've gotten some spurious bytes in the beginning from
             // keypresses, so find where the color starts.
-            const char *found = (const char *)memmem(data, len, "rgb:", 4);
-            if (found && len - (found - data) > kColorLen)  // at least 1 more
-                return found;  // Found start of color sequence and enough
-                               // bytes.
-            return nullptr;
+            const char *found = find_str(data, len, "rgb:");
+            return strchr(found, '\\') ? found : nullptr;
         });
 
     if (!start_color) return nullptr;
 
     // Assemble a standard #rrggbb string into global static buffer.
     static char result[8];
+    strcpy(result, "#000000");  // Some default.
     result[0] = '#';
-    memcpy(&result[1], &start_color[4], 2);
-    memcpy(&result[3], &start_color[9], 2);
-    memcpy(&result[5], &start_color[14], 2);
+    start_color += strlen("rgb:");
+    for (int pos = 1; pos < 6 && isxdigit(*start_color); pos += 2) {
+        memcpy(&result[pos], start_color, 2);
+        start_color = strchr(start_color, '/');  // color separator.
+        if (!start_color) return result;
+        start_color += 1;
+    }
     result[7] = '\0';
 
     return result;
 }
 
-static inline bool contains(const char *data, size_t len, const char *str) {
-    return memmem(data, len, str, strlen(str)) != nullptr;
-}
-
 TermGraphicsInfo QuerySupportedGraphicsProtocol() {
-    TermGraphicsInfo result;
-    result.preferred_graphics                  = GraphicsProtocol::kNone;
-    result.known_broken_sixel_cursor_placement = false;
+    TermGraphicsInfo result{};
+    result.preferred_graphics = GraphicsProtocol::kNone;
+    const int sixel_env_bits = timg::GetIntEnv("TIMG_SIXEL_NEWLINE_WORKAROUND");
+    result.sixel.known_broken_cursor_placement = sixel_env_bits & 0b01;
+    result.sixel.full_cell_jump                = sixel_env_bits & 0b10;
     result.in_tmux                             = false;
 
     // Environment variables can be changed, so guesses from environment
@@ -201,16 +243,29 @@ TermGraphicsInfo QuerySupportedGraphicsProtocol() {
         // Fall through, as we still have to determine if we're in tmux
     }
 
-    // vscode doesn't provide a way to query the terminal, but can do
-    // iterm graphics.
+    // Some terminals with Windows heritage often don't support query of
+    // the terminal, but are set TERM_PROGRAM
     const char *const term_program = getenv("TERM_PROGRAM");
-    if (term_program && strcmp(term_program, "vscode") == 0) {
-        result.preferred_graphics = GraphicsProtocol::kIterm2;
-        // In case the user chooses sixel.
-        result.known_broken_sixel_cursor_placement = true;
+    if (term_program) {
+        if (strcmp(term_program, "vscode") == 0) {
+            result.preferred_graphics = GraphicsProtocol::kIterm2;
+            // In case the user chooses sixel.
+            result.sixel.known_broken_cursor_placement = true;
+        }
+        else if (strcmp(term_program, "WarpTerminal") == 0) {
+            // At least on Mac and Linux according to #151
+            // (Windows support unclear)
+            result.preferred_graphics = GraphicsProtocol::kIterm2;
+        }
     }
 
+    // Note, there is a kitty protocol way to determine if kitty is supported,
+    // unfortunately that escape sequence messes up some other terminals, so
+    // we don't do that here. Instead, we match known terminals and then
+    // fall back to sixel detection.
+
     const Duration kTimeBudget = Duration::Millis(250);
+    char buffer[512];
 
     // We send out two queries: one CSI for terminal version detection that
     // is supported at least by the terminals we're interested in. From the
@@ -219,60 +274,90 @@ TermGraphicsInfo QuerySupportedGraphicsProtocol() {
     // This is followed by DSR 5 that is always expected to work.
     // If we only get a response to the innocuous status report request,
     // we don't have a terminal that supports the CSI >q
-    constexpr char term_query[] =
-        "\033[>q"   // terminal version query
-        "\033[5n";  // general status report.
-    char buffer[512];
-    QueryTerminal(term_query, buffer, sizeof(buffer), kTimeBudget,
+    constexpr char kTermVersionQuery[] =
+        TERM_CSI ">q"   // terminal version query
+        TERM_CSI "5n";  // general status report.
+    QueryTerminal(kTermVersionQuery, buffer, sizeof(buffer), kTimeBudget,
                   [&result](const char *data, size_t len) {
-                      if (contains(data, len, "iTerm2") ||
-                          contains(data, len, "Konsole 2")) {
+                      if (find_str(data, len, "iTerm2") ||
+                          find_str(data, len, "Konsole 2")) {
                           result.preferred_graphics = GraphicsProtocol::kIterm2;
                       }
-                      if (contains(data, len, "WezTerm")) {
+                      if (find_str(data, len, "WezTerm")) {
                           result.preferred_graphics = GraphicsProtocol::kIterm2;
-                          result.known_broken_sixel_cursor_placement = true;
+                          result.sixel.known_broken_cursor_placement = true;
                       }
-                      if (contains(data, len, "kitty")) {
+                      if (find_str(data, len, "kitty")) {
                           result.preferred_graphics = GraphicsProtocol::kKitty;
                       }
-                      if (contains(data, len, "ghostty")) {
+                      if (find_str(data, len, "ghostty")) {
                           result.preferred_graphics = GraphicsProtocol::kKitty;
                       }
-                      if (contains(data, len, "mlterm")) {
+                      if (find_str(data, len, "mlterm")) {
                           result.preferred_graphics = GraphicsProtocol::kSixel;
                       }
-                      if (contains(data, len, "XTerm")) {
-                          result.preferred_graphics = GraphicsProtocol::kSixel;
-                          result.known_broken_sixel_cursor_placement = true;
+                      if (find_str(data, len, "XTerm")) {
+                          // Don't know yet if supports sixel, will query below
+                          result.sixel.known_broken_cursor_placement = true;
                       }
-                      if (contains(data, len, "foot")) {
+                      if (find_str(data, len, "foot")) {
                           result.preferred_graphics = GraphicsProtocol::kSixel;
-                          result.known_broken_sixel_cursor_placement = true;
+                          result.sixel.known_broken_cursor_placement = true;
                       }
-                      if (contains(data, len, "tmux")) {
+                      if (find_str(data, len, "tmux")) {
                           result.in_tmux = true;
                       }
+                      if (find_str(data, len, "WindowsTerminal")) {
+                          // TODO: check again once name is established
+                          // https://github.com/microsoft/terminal/issues/18382
+                          result.sixel.known_broken_cursor_placement = true;
+                          result.sixel.full_cell_jump                = true;
+                      }
                       // We finish once we found the response to DSR5
-                      return (const char *)memmem(data, len, "\033[0n", 3);
+                      return find_str(data, len, TERM_CSI "0");
+                  });
+    if (result.preferred_graphics != GraphicsProtocol::kNone) {
+        return result;
+    }
+
+    // Still not known. Let's see if if we can at determine if this might
+    // be sixel as some terminals implement DA1 (primary device attributes)
+    // with ";4" in its response if it supports sixel.
+    QueryTerminal(TERM_CSI "c", buffer, sizeof(buffer), kTimeBudget,
+                  [&result](const char *data, size_t len) {
+                      // https://vt100.net/docs/vt510-rm/DA1.html
+                      // says CSI ?64 is returned, but not all terminals do
+                      // that. So, let's just watch for CSI ?
+                      const char *const end    = data + len;
+                      constexpr char kExpect[] = TERM_CSI "?";
+                      const char *start        = find_str(data, len, kExpect);
+                      if (!start) return start;
+                      // Look for length - 1 as we expect another char we test
+                      // below.
+                      const char *found_4 =
+                          find_str(start, end - start - 1, ";4");
+                      if (found_4 && (found_4[2] == ';' || found_4[2] == 'c')) {
+                          result.preferred_graphics = GraphicsProtocol::kSixel;
+                      }
+                      return data;
                   });
     return result;
 }
 
 static bool QueryCellWidthHeight(int *width, int *height) {
     const Duration kTimeBudget      = Duration::Millis(50);
-    constexpr char query[]          = "\033[16t";
-    constexpr char response_start[] = "\033[6;";
+    constexpr char kQuery[]         = TERM_CSI "16t";
+    constexpr char kResponseStart[] = TERM_CSI "6;";
     char buffer[512];
     const char *const result = QueryTerminal(
-        query, buffer, sizeof(buffer), kTimeBudget,
-        [response_start](const char *data, size_t len) -> const char * {
-            return (const char *)memmem(data, len, response_start,
-                                        strlen(response_start));
+        kQuery, buffer, sizeof(buffer), kTimeBudget,
+        [kResponseStart](const char *data, size_t len) -> const char * {
+            return find_str(data, len, kResponseStart);
         });
-    int w, h;
+    int w;
+    int h;
     if (!result ||
-        sscanf(result + strlen(response_start), "%d;%dt", &h, &w) != 2) {
+        sscanf(result + strlen(kResponseStart), "%d;%dt", &h, &w) != 2) {
         return false;
     }
     *width  = w;
@@ -283,9 +368,14 @@ static bool QueryCellWidthHeight(int *width, int *height) {
 // Probe all file descriptors that might be connect to tty for term size.
 TermSizeResult DetermineTermSize() {
     TermSizeResult result;
-    for (int fd : {STDOUT_FILENO, STDERR_FILENO, STDIN_FILENO}) {
+    for (const int fd : {STDOUT_FILENO, STDERR_FILENO, STDIN_FILENO}) {
         struct winsize w = {};
-        if (ioctl(fd, TIOCGWINSZ, &w) != 0) continue;
+        if (ioctl(fd, TIOCGWINSZ, &w) != 0) {
+            if (s_log_terminal_queries) {
+                fprintf(stderr, "ioctl(%d, TIOCGWINSZ) failing.\n", fd);
+            }
+            continue;
+        }
         // If we get the size of the terminals in pixels, we can determine
         // what aspect ratio the pixels. This is also needed to
         // jump up the exact number of character cells needed for animations
@@ -301,6 +391,9 @@ TermSizeResult DetermineTermSize() {
             result.font_height_px = w.ws_ypixel / w.ws_row;
         }
         else {
+            if (s_log_terminal_queries) {
+                fprintf(stderr, "No usable TIOCGWINSZ, trying cell query.\n");
+            }
             // Alright, TIOCGWINSZ did not return the terminal size, let's
             // see if it reports character cell size otherwise
             QueryCellWidthHeight(&result.font_width_px, &result.font_height_px);
